use crate::speed_unit::speed_to_unit_string;
use crate::temp_unit::temp_to_unit_string;
use crate::{
    coords::Coordinates, current_weather::CurrentWeather,
    current_weather_print_params::CurrentWeatherPrintParams,
};

pub struct CurrentWeatherPrinter {
    current_weather: CurrentWeather,
    params: CurrentWeatherPrintParams,
    coords: Coordinates,
    city: Option<String>,
}

impl CurrentWeatherPrinter {
    pub fn new(
        current_weather: CurrentWeather,
        params: CurrentWeatherPrintParams,
        coords: Coordinates,
        city: Option<String>,
    ) -> CurrentWeatherPrinter {
        CurrentWeatherPrinter {
            current_weather,
            params,
            coords,
            city,
        }
    }

    pub fn extract_string(&self) -> String {
        let mut string_vec: Vec<String> = Vec::new();

        if !self.params.clean {
            string_vec.push(self.create_header());
        } else {
            if self.params.all || self.params.include_coords {
                string_vec.push(self.coords.latitude.to_string());
                string_vec.push(self.coords.longitude.to_string());
            }
            if (self.params.all || self.params.include_city) && self.city.is_some() {
                string_vec.push(self.city.clone().unwrap());
            }
        }

        if self.params.is_day || self.params.all {
            string_vec.push(self.extract_day());
        }
        if self.params.temperature || self.params.all {
            string_vec.push(self.extract_temperature());
        }
        if self.params.windspeed || self.params.all {
            string_vec.push(self.extract_wind_speed());
        }
        if self.params.winddirection || self.params.all {
            string_vec.push(self.extract_wind_direction());
        }

        if !self.params.clean && (self.params.all || self.params.include_coords) {
            string_vec.push(format!(
                "{}, {}",
                self.parse_simple_data(&self.coords.latitude.to_string(), "Latitude", None),
                self.parse_simple_data(&self.coords.longitude.to_string(), "Longitude", None)
            ));
        }

        if self.params.clean {
            let final_string = string_vec.join(",");
            final_string
        } else {
            string_vec.push(self.create_footer());
            string_vec.join("\n")
        }
    }

    fn create_header(&self) -> String {
        let mut title_header: Vec<String> = Vec::new();
        title_header.push(String::from("=== Current weather"));

        if (self.params.all || self.params.include_city) && self.city.is_some() {
            title_header.push(String::from("for"));
            title_header.push(self.city.clone().unwrap());
        }

        title_header.push(String::from("==="));
        title_header.join(" ")
    }

    fn create_footer(&self) -> String {
        String::from("=== Weather data by Open-Meteo.com ===")
    }

    fn extract_day(&self) -> String {
        if self.current_weather.is_day == 1 {
            self.parse_custom_data(&self.current_weather.is_day.to_string(), "Day")
        } else {
            self.parse_custom_data(&self.current_weather.is_day.to_string(), "Night")
        }
    }

    fn extract_temperature(&self) -> String {
        self.parse_simple_data(
            &self.current_weather.temperature.to_string(),
            "Temperature",
            Some(temp_to_unit_string(&self.params.temperature_unit).as_str()),
        )
    }

    fn extract_wind_speed(&self) -> String {
        self.parse_simple_data(
            &self.current_weather.windspeed.to_string(),
            "Wind speed",
            Some(format!(" {}", speed_to_unit_string(&self.params.speed_unit)).as_str()),
        )
    }

    fn extract_wind_direction(&self) -> String {
        self.parse_simple_data(
            &self.current_weather.winddirection.to_string(),
            "Wind direction",
            Some("°"),
        )
    }

    fn parse_custom_data(&self, data: &str, custom: &str) -> String {
        if self.params.clean {
            format!("{data}")
        } else {
            format!("{custom}")
        }
    }

    fn parse_simple_data(&self, data: &str, descriptor: &str, end_text: Option<&str>) -> String {
        if self.params.clean {
            format!("{data}")
        } else {
            let end_text = end_text.unwrap_or("");
            format!("{descriptor}: {data}{end_text}")
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::speed_unit::SpeedUnit;
    use crate::temp_unit::TempUnit;

    #[test]
    fn clean_all_data() {
        let current_weather = CurrentWeather {
            is_day: 1,
            temperature: 12.5,
            windspeed: 7.0,
            winddirection: 90.0,
        };
        let params = CurrentWeatherPrintParams::new(
            true,
            false,
            false,
            TempUnit::Celsius,
            false,
            SpeedUnit::Kmh,
            false,
            false,
            false,
            true,
        );
        let coords = Coordinates::new(5.0, -5.0);
        let printer = CurrentWeatherPrinter::new(
            current_weather,
            params,
            coords,
            Some(String::from("TestCity")),
        );

        assert_eq!(printer.extract_string(), "5,-5,TestCity,1,12.5,7,90");
    }

    #[test]
    fn clean_basic_data() {
        let current_weather = CurrentWeather {
            is_day: 1,
            temperature: 15.5,
            windspeed: 12.2,
            winddirection: 150.0,
        };
        let params = CurrentWeatherPrintParams::new(
            false,
            true,
            true,
            TempUnit::Celsius,
            true,
            SpeedUnit::Kmh,
            false,
            false,
            false,
            true,
        );
        let coords = Coordinates::new(12.0, -55.0);
        let printer = CurrentWeatherPrinter::new(
            current_weather,
            params,
            coords,
            Some(String::from("TestCity")),
        );

        assert_eq!(printer.extract_string(), "1,15.5,12.2");
    }

    #[test]
    fn full_data() {
        let current_weather = CurrentWeather {
            is_day: 0,
            temperature: 22.0,
            windspeed: 15.5,
            winddirection: 118.0,
        };
        let params = CurrentWeatherPrintParams::new(
            true,
            false,
            false,
            TempUnit::Celsius,
            false,
            SpeedUnit::Kmh,
            false,
            false,
            false,
            false,
        );
        let coords = Coordinates::new(5.0, -5.0);
        let printer = CurrentWeatherPrinter::new(
            current_weather,
            params,
            coords,
            Some(String::from("TestCity")),
        );

        let output = printer.extract_string();
        assert!(output.contains("Night"));
        assert!(output.contains("Temperature: 22°C"));
        assert!(output.contains("Wind speed: 15.5 km/h"));
        assert!(output.contains("Wind direction: 118°"));
        assert!(output.contains("Latitude: 5"));
        assert!(output.contains("Longitude: -5"));
        assert!(output.contains("TestCity"));
        assert!(output.contains("Open-Meteo.com"));
    }

    #[test]
    fn full_basic_data() {
        let current_weather = CurrentWeather {
            is_day: 1,
            temperature: 55.0,
            windspeed: 11.5,
            winddirection: 125.0,
        };
        let params = CurrentWeatherPrintParams::new(
            false,
            true,
            true,
            TempUnit::Fahrenheit,
            true,
            SpeedUnit::Mph,
            false,
            false,
            false,
            false,
        );
        let coords = Coordinates::new(12.15, 0.235);
        let printer = CurrentWeatherPrinter::new(
            current_weather,
            params,
            coords,
            Some(String::from("NoCity")),
        );

        let output = printer.extract_string();
        assert!(output.contains("Day"));
        assert!(output.contains("Temperature: 55°F"));
        assert!(output.contains("Wind speed: 11.5 mp/h"));
        assert!(!output.contains("Wind direction: 125°"));
        assert!(!output.contains("Latitude: 12.15"));
        assert!(!output.contains("Longitude: 0.235"));
        assert!(!output.contains("Nocity"));
        assert!(output.contains("Open-Meteo.com"));
    }
}
